@@ -1,25 +1,57 @@
 # Functions for managing BIND 4 and 8/9 records files
+use strict;
+use warnings;
+no warnings 'redefine';
+no warnings 'uninitialized';
+
+# Globals from Webmin or bind8-lib.pl
+our (%config, %text, %in);
+our $module_config_directory;
+our $bind_version;
+our $ipv6revzone = $config{'ipv6_mode'} ? "ip6.arpa" : "ip6.int";
 
 # read_zone_file(file, origin, [previous], [only-soa], [no-chroot])
 # Reads a DNS zone file and returns a data structure of records. The origin
 # must be a domain without the trailing dot, or just .
 sub read_zone_file
 {
-local($file, $lnum, $line, $t, @tok, @lnum, @coms,
-      $i, @rv, $origin, $num, $j, @inc, @oset, $comment);
+my ($file, $line, @tok, @lnum, @coms,
+      @rv, $origin, @inc, @oset, $comment);
 $origin = $_[1];
+if (&has_ndc() == 2) {
+	# Flush the zone file
+	&backquote_command(
+		$config{'rndc_cmd'}.
+		($config{'rndc_conf'} ? " -c $config{'rndc_conf'}" : "").
+		" sync ".quotemeta($origin)." 2>&1 </dev/null");
+	}
 if ($origin ne ".") {
 	# Remove trailing dots in origin name, as they are added automatically
 	# in the code below.
 	$origin =~ s/\.*$//;
 	}
 $file = &absolute_path($_[0]);
-local $rootfile = $_[4] ? $file : &make_chroot($file);
-open(FILE, $rootfile);
-$lnum = 0;
-local ($gotsoa, $aftersoa);
-while($line = <FILE>) {
-	local($glen, $merged_2, $merge);
+my $rootfile = $_[4] ? $file : &make_chroot($file);
+my $FILE;
+if (&is_raw_format_records($rootfile)) {
+	# Convert from raw format first
+	my $compilezone = &has_command("named-compilezone");
+	$compilezone ||
+		&error("Zone file $rootfile is in raw format, but the ".
+		       "named-compilezone command is not installed");
+	open($FILE, "-|", $compilezone,
+	     "-f", "raw", "-F", "text", "-o", "-",
+	     $origin, $rootfile);
+	}
+else {
+	# Can read text format records directly
+	open($FILE, "<", $rootfile);
+	}
+my $lnum = 0;
+my ($gotsoa, $aftersoa) = (0, 0);
+while($line = <$FILE>) {
+	my ($glen, $merged_2, $merge);
+	$glen = 0;
 	# strip comments (# is not a valid comment separator here!)
 	$line =~ s/\r|\n//g;
 	# parsing splited into separate cases to fasten it
@@ -45,10 +77,10 @@ while($line = <FILE>) {
 		}
 
 	# split line into tokens
-	local $oset = 0;
+	my $oset = 0;
 	while(1) {
 		$merge = 1;
-		$base_oset = 0;
+		my $base_oset = 0;
 		if ($line =~ /^(\s*)\"((?:[^\"\\]|\\.)*)\"(.*)/ ||
 		    $line =~ /^(\s*)((?:[^\s\(\)\"\\]|\\.)+)(.*)/ ||
 		    ($merge = 0) || $line =~ /^(\s*)([\(\)])(.*)/) {
@@ -88,10 +120,12 @@ while($line = <FILE>) {
 		last;
 		}
 	}
-close(FILE);
+close($FILE);
 
 # parse into data structures
-$i = 0; $num = 0;
+my $i = 0;
+my $num = 0;
+my $currttl = undef;
 while($i < @tok) {
 	if ($tok[$i] =~ /^\$origin$/i) {
 		# $ORIGIN directive (may be relative or absolute)
@@ -106,7 +140,7 @@ while($i < @tok) {
 		# including another file
 		if ($lnum[$i+1] == $lnum[$i+2]) {
 			# $INCLUDE zonefile origin
-			local $inc_origin;
+			my $inc_origin;
 			if ($tok[$i+2] =~ /^(\S+)\.$/) {
 				$inc_origin = $1 ? $1 : ".";
 				}
@@ -122,17 +156,18 @@ while($i < @tok) {
 					       @rv ? $rv[$#rv] : undef);
 			$i += 2;
 			}
-		foreach $j (@inc) { $j->{'num'} = $num++; }
+		foreach my $j (@inc) { $j->{'num'} = $num++; }
 		push(@rv, @inc);
 		}
 	elsif ($tok[$i] =~ /^\$generate$/i) {
 		# a generate directive .. add it as a special record
-		local $gen = { 'file' => $file,
-			       'rootfile' => $rootfile,
-			       'comment' => $coms[$i],
-			       'line' => $lnum[$i],
-			       'num' => $num++ };
-		local @gv;
+		my $gen = { 'file' => $file,
+			    'rootfile' => $rootfile,
+			    'comment' => $coms[$i],
+			    'line' => $lnum[$i],
+			    'num' => $num++,
+			    'type' => '' };
+		my @gv;
 		while($lnum[++$i] == $gen->{'line'}) {
 			push(@gv, $tok[$i]);
 			}
@@ -142,23 +177,25 @@ while($i < @tok) {
 	elsif ($tok[$i] =~ /^\$ttl$/i) {
 		# a ttl directive
 		$i++;
-		local $defttl = { 'file' => $file,
-				  'rootfile' => $rootfile,
-			       	  'line' => $lnum[$i],
-			          'num' => $num++,
-			       	  'defttl' => $tok[$i++] };
+		my $defttl = { 'file' => $file,
+			       'rootfile' => $rootfile,
+		      	       'line' => $lnum[$i],
+		               'num' => $num++,
+		       	       'defttl' => $tok[$i++],
+			       'type' => '' };
 		push(@rv, $defttl);
+		$currttl = $defttl->{'defttl'};
 		}
 	elsif ($tok[$i] =~ /^\$(\S+)/i) {
 		# some other special directive
-		local $ln = $lnum[$i];
+		my $ln = $lnum[$i];
 		while($lnum[$i] == $ln) {
 			$i++;
 			}
 		}
 	else {
 		# A DNS record line
-		local(%dir, @values, $l);
+		my(%dir, @values, $l);
 		$dir{'line'} = $lnum[$i];
 		$dir{'file'} = $file;
 		$dir{'rootfile'} = $rootfile;
@@ -211,7 +248,8 @@ while($i < @tok) {
 			$dir{'class'} = "IN";
 			$i++;
 			}
-		if ($dir{'name'} eq '') {
+		if (!defined($dir{'name'}) || $dir{'name'} eq '') {
+			my $prv;
 			# Name comes from previous record
 			for(my $p=$#rv; $p>=0; $p--) {
 				$prv = $rv[$p];
@@ -230,14 +268,14 @@ while($i < @tok) {
 		# read values until end of line, unless a ( is found, in which
 		# case read till the )
 		$l = $lnum[$i];
-		while($lnum[$i] == $l && $i < @tok) {
+		while($i < @tok && $lnum[$i] == $l) {
 			if ($tok[$i] eq "(") {
 				my $olnum = $lnum[$i];
 				while($tok[++$i] ne ")") {
 					push(@values, $tok[$i]);
 					if ($i >= @tok) {
 						&error("No ending ) found for ".
-						       "( starting at $olnum");
+						       "( at $olnum in $file");
 						}
 					}
 				$i++; # skip )
@@ -247,6 +285,7 @@ while($i < @tok) {
 			}
 		$dir{'values'} = \@values;
 		$dir{'eline'} = $lnum[$i-1];
+		$dir{'realttl'} = $dir{'ttl'} || $currttl;
 
 		# Work out canonical form, and maybe use it
 		my $canon = $dir{'name'};
@@ -263,14 +302,24 @@ while($i < @tok) {
 		$dir{'num'} = $num++;
 
 		# If this is an SPF record .. adjust the class
-		local $spf;
+		my $spf;
 		if ($dir{'type'} eq 'TXT' &&
+		    !$config{'spf_record'} &&
 		    ($spf=&parse_spf(@{$dir{'values'}}))) {
-			if (!@{$spf->{'other'}}) {
+			if (!$spf->{'other'} || !@{$spf->{'other'}}) {
 				$dir{'type'} = 'SPF';
 				}
 			}
 
+		# If this is a DMARC record .. adjust the class
+		my $dmarc;
+		if ($dir{'type'} eq 'TXT' &&
+                    ($dmarc=&parse_dmarc(@{$dir{'values'}}))) {
+                        if (!$dmarc->{'other'} || !@{$dmarc->{'other'}}) {
+                                $dir{'type'} = 'DMARC';
+                                }
+                        }
+
 		push(@rv, \%dir);
 
 		# Stop processing if this was an SOA record
@@ -282,13 +331,49 @@ while($i < @tok) {
 return @rv;
 }
 
+# files_in_zone_file(file)
+# Quickly finds all includes in a zone file
+sub files_in_zone_file
+{
+my ($file) = @_;
+my @rv = ( $file );
+my $fh;
+open($fh, "<", $file);
+while(<$fh>) {
+	if (/^\$include\s+(\S+)/) {
+		my $inc = $1;
+		push(@rv, &files_in_zone_file($inc));
+		}
+	}
+close($fh);
+return @rv;
+}
+
 # create_record(file, name, ttl, class, type, values, comment)
 # Add a new record of some type to some zone file
 sub create_record
 {
-local $fn = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($fn);
-push(@$lref, &make_record(@_[1..$#_]));
+my ($file, @rec) = @_;
+my $fn = &make_chroot(&absolute_path($file));
+&is_raw_format_records($fn) && &error("Raw format zone files cannot be edited");
+my $lref = &read_file_lines($fn);
+push(@$lref, &make_record(@rec));
+&flush_file_lines($fn);
+}
+
+# create_multiple_records(file, &records)
+# Create records from structures 
+sub create_multiple_records
+{
+my ($file, $recs) = @_;
+my $fn = &make_chroot(&absolute_path($file));
+&is_raw_format_records($fn) && &error("Raw format zone files cannot be edited");
+my $lref = &read_file_lines($fn);
+foreach my $r (@$recs) {
+	push(@$lref, &make_record($r->{'name'}, $r->{'ttl'}, $r->{'class'},
+			          $r->{'type'}, join(" ", @{$r->{'values'}}),
+			          $r->{'comment'}));
+	}
 &flush_file_lines($fn);
 }
 
@@ -296,9 +381,10 @@ push(@$lref, &make_record(@_[1..$#_]));
 # Updates an existing record in some zone file
 sub modify_record
 {
-local $fn = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($fn);
-local $lines = $_[1]->{'eline'} - $_[1]->{'line'} + 1;
+my $fn = &make_chroot(&absolute_path($_[0]));
+&is_raw_format_records($fn) && &error("Raw format zone files cannot be edited");
+my $lref = &read_file_lines($fn);
+my $lines = $_[1]->{'eline'} - $_[1]->{'line'} + 1;
 splice(@$lref, $_[1]->{'line'}, $lines, &make_record(@_[2..$#_]));
 &flush_file_lines($fn);
 }
@@ -307,10 +393,27 @@ splice(@$lref, $_[1]->{'line'}, $lines, &make_record(@_[2..$#_]));
 # Deletes a record in some zone file
 sub delete_record
 {
-local $fn = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($fn);
-local $lines = $_[1]->{'eline'} - $_[1]->{'line'} + 1;
-splice(@$lref, $_[1]->{'line'}, $lines);
+my ($file, $r) = @_;
+my $fn = &make_chroot(&absolute_path($file));
+&is_raw_format_records($fn) && &error("Raw format zone files cannot be edited");
+my $lref = &read_file_lines($fn);
+my $lines = $r->{'eline'} - $r->{'line'} + 1;
+splice(@$lref, $r->{'line'}, $lines);
+&flush_file_lines($fn);
+}
+
+# delete_multiple_records(file, &records)
+# Delete many records from the same file at once
+sub delete_multiple_records
+{
+my ($file, $recs) = @_;
+my $fn = &make_chroot(&absolute_path($file));
+&is_raw_format_records($fn) && &error("Raw format zone files cannot be edited");
+my $lref = &read_file_lines($fn);
+foreach my $r (sort { $b->{'line'} <=> $a->{'line'} } @$recs) {
+	my $lines = $r->{'eline'} - $r->{'line'} + 1;
+	splice(@$lref, $r->{'line'}, $lines);
+	}
 &flush_file_lines($fn);
 }
 
@@ -318,8 +421,8 @@ splice(@$lref, $_[1]->{'line'}, $lines);
 # Add a new $generate line to some zone file
 sub create_generator
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 push(@$lref, join(" ", '$generate', @_[1..4]).
 	     ($_[5] ? " ;$_[5]" : ""));
 &flush_file_lines($f);
@@ -329,8 +432,8 @@ push(@$lref, join(" ", '$generate', @_[1..4]).
 # Updates an existing $generate line in some zone file
 sub modify_generator
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 $lref->[$_[1]->{'line'}] = join(" ", '$generate', @_[2..5]).
 			   ($_[6] ? " ;$_[6]" : "");
 &flush_file_lines($f);
@@ -340,8 +443,8 @@ $lref->[$_[1]->{'line'}] = join(" ", '$generate', @_[2..5]).
 # Deletes a $generate line in some zone file
 sub delete_generator
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 splice(@$lref, $_[1]->{'line'}, 1);
 &flush_file_lines($f);
 }
@@ -350,8 +453,8 @@ splice(@$lref, $_[1]->{'line'}, 1);
 # Adds a $ttl line to a records file
 sub create_defttl
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 splice(@$lref, 0, 0, "\$ttl $_[1]");
 &flush_file_lines($f);
 }
@@ -360,8 +463,8 @@ splice(@$lref, 0, 0, "\$ttl $_[1]");
 # Updates the $ttl line with a new value
 sub modify_defttl
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 $lref->[$_[1]->{'line'}] = "\$ttl $_[2]";
 &flush_file_lines($f);
 }
@@ -370,43 +473,47 @@ $lref->[$_[1]->{'line'}] = "\$ttl $_[2]";
 # Removes the $ttl line from a records file
 sub delete_defttl
 {
-local $f = &make_chroot(&absolute_path($_[0]));
-local $lref = &read_file_lines($f);
+my $f = &make_chroot(&absolute_path($_[0]));
+my $lref = &read_file_lines($f);
 splice(@$lref, $_[1]->{'line'}, 1);
 &flush_file_lines($f);
 }
 
-
-
 # make_record(name, ttl, class, type, values, comment)
 # Returns a string for some zone record
 sub make_record
 {
-local $type = $_[3] eq "SPF" ? "TXT" : $_[3];
-return $_[0] . ($_[1] ? "\t$_[1]" : "") . "\t$_[2]\t$type\t$_[4]" .
-       ($_[5] ? "\t;$_[5]" : "");
+my ($name, $ttl, $cls, $type, $values, $cmt) = @_;
+$type = $type eq "SPF" && !$config{'spf_record'} ? "TXT" :
+        $type eq "DMARC" ? "TXT" : $type;
+return $name.
+       (defined($ttl) && $ttl ne "" ? "\t".$ttl : "").
+       "\t".$cls.
+       "\t".$type.
+       "\t".$values.
+       (defined($cmt) && $cmt ne "" ? "\t;$cmt" : "");
 }
 
 # bump_soa_record(file, &records)
 # Increase the serial number in some SOA record by 1
 sub bump_soa_record
 {
-local($i, $r, $v, $vals);
-for($i=0; $i<@{$_[1]}; $i++) {
-	$r = $_[1]->[$i];
+my ($file, $recs) = @_;
+foreach my $r (@$recs) {
 	if ($r->{'type'} eq "SOA") {
-		$v = $r->{'values'};
+		my $v = $r->{'values'};
 		# already set serial if no acl allow it to update or update
 		# is disabled
-		$serial = $v->[2];
+		my $serial = $v->[2];
 		if ($config{'updserial_on'}) {
 			# automatically handle serial numbers ?
 			$serial = &compute_serial($v->[2]);
 			}
-		$vals = "$v->[0] $v->[1] (\n\t\t\t$serial\n\t\t\t$v->[3]\n".
-			"\t\t\t$v->[4]\n\t\t\t$v->[5]\n\t\t\t$v->[6] )";
+		my $vals = "$v->[0] $v->[1] (\n\t\t\t$serial\n\t\t\t$v->[3]\n".
+			   "\t\t\t$v->[4]\n\t\t\t$v->[5]\n\t\t\t$v->[6] )";
 		&modify_record($r->{'file'}, $r, $r->{'realname'}, $r->{'ttl'},
 				$r->{'class'}, $r->{'type'}, $vals);
+		$v->[2] = $serial;
 		}
 	}
 }
@@ -415,25 +522,30 @@ for($i=0; $i<@{$_[1]}; $i++) {
 # Returns a string like YYYYMMDD
 sub date_serial
 {
-local $now = time();
-local @tm = localtime($now);
+my $now = time();
+my @tm = localtime($now);
 return sprintf "%4.4d%2.2d%2.2d", $tm[5]+1900, $tm[4]+1, $tm[3];
 }
 
 # get_zone_defaults(&hash)
 sub get_zone_defaults
 {
-if (!&read_file("$module_config_directory/zonedef", $_[0])) {
-	$_[0]->{'refresh'} = 10800; $_[0]->{'retry'} = 3600;
-	$_[0]->{'expiry'} = 604800; $_[0]->{'minimum'} = 38400;
-	$_[0]->{'refunit'} = ""; $_[0]->{'retunit'} = "";
-	$_[0]->{'expunit'} = ""; $_[0]->{'minunit'} = "";
+my ($zd) = @_;
+if (!&read_file("$module_config_directory/zonedef", $zd)) {
+	$zd->{'refresh'} = 3600;
+	$zd->{'retry'} = 600;
+	$zd->{'expiry'} = 1209600;
+	$zd->{'minimum'} = 3600;
+	$zd->{'refunit'} = "";
+	$zd->{'retunit'} = "";
+	$zd->{'expunit'} = "";
+	$zd->{'minunit'} = "";
 	}
 else {
-	$_[0]->{'refunit'} = $1 if ($_[0]->{'refresh'} =~ s/([^0-9])$//);
-	$_[0]->{'retunit'} = $1 if ($_[0]->{'retry'} =~ s/([^0-9])$//);
-	$_[0]->{'expunit'} = $1 if ($_[0]->{'expiry'} =~ s/([^0-9])$//);
-	$_[0]->{'minunit'} = $1 if ($_[0]->{'minimum'} =~ s/([^0-9])$//);
+	$zd->{'refunit'} = $1 if ($zd->{'refresh'} =~ s/([^0-9])$//);
+	$zd->{'retunit'} = $1 if ($zd->{'retry'} =~ s/([^0-9])$//);
+	$zd->{'expunit'} = $1 if ($zd->{'expiry'} =~ s/([^0-9])$//);
+	$zd->{'minunit'} = $1 if ($zd->{'minimum'} =~ s/([^0-9])$//);
 	}
 }
 
@@ -448,7 +560,7 @@ sub allowed_zone_file
 {
 return 0 if ($_[1] =~ /\.\./);
 return 0 if (-l $_[1] && !&allowed_zone_file($_[0], readlink($_[1])));
-local $l = length($_[0]->{'dir'});
+my $l = length($_[0]->{'dir'});
 return length($_[1]) > $l && substr($_[1], 0, $l) eq $_[0]->{'dir'};
 }
 
@@ -456,50 +568,61 @@ return length($_[1]) > $l && substr($_[1], 0, $l) eq $_[0]->{'dir'};
 sub sort_records
 {
 return @_ if (!@_);
-local $s = $in{'sort'} ? $in{'sort'} : $config{'records_order'};
+my $s = $in{'sort'} ? $in{'sort'} : $config{'records_order'};
 if ($s == 1) {
 	# Sort by name
 	if ($_[0]->{'type'} eq "PTR") {
-		return sort ptr_sort_func @_;
+		my @rv = sort ptr_sort_func @_; 
+		return @rv;
 		}
 	else {
-		return sort { $a->{'name'} cmp $b->{'name'} } @_;
+		my @rv = sort { $a->{'name'} cmp $b->{'name'} } @_;
+		return @rv;
 		}
 	}
 elsif ($s == 2) {
 	# Sort by value
 	if ($_[0]->{'type'} eq "A") {
-		return sort ip_sort_func @_;
+		my @rv = sort ip_sort_func @_;
+		return @rv;
 		}
 	elsif ($_[0]->{'type'} eq "MX") {
-		return sort { $a->{'values'}->[1] cmp $b->{'values'}->[1] } @_;
+		my @rv = sort { $a->{'values'}->[1] cmp $b->{'values'}->[1] } @_;
+		return @rv;
 		}
 	else {
-		return sort { $a->{'values'}->[0] cmp $b->{'values'}->[0] } @_;
+		my @rv = sort { $a->{'values'}->[0] cmp $b->{'values'}->[0] } @_;
+		return @rv;
 		}
 	}
 elsif ($s == 3) {
 	# Sort by IP address or by value if there is no IP
 	if ($_[0]->{'type'} eq "A") {
-		return sort ip_sort_func @_;
+		my @rv = sort ip_sort_func @_;
+		return @rv;
 		}
 	elsif ($_[0]->{'type'} eq "PTR") {
-		return sort ptr_sort_func @_;
+		my @rv = sort ptr_sort_func @_;
+		return @rv;
 		}
 	elsif ($_[0]->{'type'} eq "MX") {
-		return sort { $a->{'values'}->[1] cmp $b->{'values'}->[1] } @_;
+		my @rv = sort { $a->{'values'}->[1] cmp $b->{'values'}->[1] } @_;
+		return @rv;
 		}
 	else {
-		return sort { $a->{'values'}->[0] cmp $b->{'values'}->[0] } @_;
+		my @rv = sort { $a->{'values'}->[0] cmp $b->{'values'}->[0] } @_;
+		return @rv;
 		}
 	}
 elsif ($s == 4) {
 	# Sort by comment
-	return sort { $b->{'comment'} cmp $a->{'comment'} } @_;
+	my @rv = sort { $b->{'comment'} cmp $a->{'comment'} } @_;
+	return @rv;
 	}
 elsif ($s == 5) {
 	# Sort by type
-	return sort { $a->{'type'} cmp $b->{'type'} } @_;
+	my @rv = sort { $a->{'type'} cmp $b->{'type'} } @_;
+	return @rv;
 	}
 else {
 	return @_;
@@ -509,7 +632,7 @@ else {
 sub ptr_sort_func
 {
 $a->{'name'} =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)/;
-local ($a1, $a2, $a3, $a4) = ($1, $2, $3, $4);
+my ($a1, $a2, $a3, $a4) = ($1, $2, $3, $4);
 $b->{'name'} =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)/;
 return	$a4 < $4 ? -1 :
 	$a4 > $4 ? 1 :
@@ -524,7 +647,7 @@ return	$a4 < $4 ? -1 :
 sub ip_sort_func
 {
 $a->{'values'}->[0] =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)/;
-local ($a1, $a2, $a3, $a4) = ($1, $2, $3, $4);
+my ($a1, $a2, $a3, $a4) = ($1, $2, $3, $4);
 $b->{'values'}->[0] =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)/;
 return	$a1 < $1 ? -1 :
 	$a1 > $1 ? 1 :
@@ -556,13 +679,12 @@ if ($_[0] =~ /^([\d\-\.\/]+)$/) {
 return $_[0];
 }
 
-$ipv6revzone = $config{'ipv6_mode'} ? "ip6.arpa" : "ip6.int";
-
 # ip6int_to_net(name)
 # Converts an address like a.b.c.d.4.3.2.1.ip6.int. to 1234:dcba::
 sub ip6int_to_net
 {
-local($n, $addr = $_[0]);
+my $n;
+my $addr = $_[0];
 if ($addr =~ /^([\da-f]\.)+$ipv6revzone/i) {
 	$addr =~ s/\.$ipv6revzone/\./i;
 	$addr = reverse(split(/\./, $addr));
@@ -584,11 +706,12 @@ if ($addr =~ /^([\da-f]\.)+$ipv6revzone/i) {
 return $_[0];
 }
 
-# net_to_ip6int(address, bits)
+# net_to_ip6int(address, [bits])
 # Converts an IPv6 address like 1234:dcba:: to a.b.c.d.4.3.2.1.ip6.int.
 sub net_to_ip6int
 {
-local($addr = lc($_[0]), $n = $_[1] >> 2);
+my $addr = lc($_[0]);
+my $n = $_[1] ? $_[1] >> 2 : 0;
 if (&check_ip6address($addr)) {
 	$addr = reverse(split(/\:/, &expandall_ip6($addr)));
 	$addr =~ s/(\w)/$1\./g;
@@ -600,14 +723,15 @@ if (&check_ip6address($addr)) {
 return $addr;
 }
 
-$uscore = $config{'allow_underscore'} ? "_" : "";
-$star = $config{'allow_wild'} ? "\\*" : "";
+our $uscore = $config{'allow_underscore'} ? "_" : "";
+our $star = $config{'allow_wild'} ? "\\*" : "";
 
 # valdnsname(name, wild, origin)
+# Returns 1 if a DNS record has a valid name
 sub valdnsname
 {
-local($fqdn);
-$fqdn = $_[0] !~ /\.$/ ? "$_[0].$_[2]." : $_[0];
+my ($name, $wild, $origin) = @_;
+my $fqdn = $name !~ /\.$/ ? "$name.$origin." : $name;
 if (length($fqdn) > 255) {
 	&error(&text('edit_efqdn', $fqdn));
 	}
@@ -615,19 +739,19 @@ if ($_[0] =~ /[^\.]{64}/) {
 	# no label longer than 63 chars
 	&error(&text('edit_elabel', $_[0]));
 	}
-return ((($_[1] && $config{'allow_wild'})
-	 ? (($_[0] =~ /^[\*A-Za-z0-9\-\.$uscore]+$/)
-	   && ($_[0] !~ /.\*/ || $bind_version >= 9) # "*" can be only the first
+return ((($wild && $config{'allow_wild'})
+	 ? (($name =~ /^[\*A-Za-z0-9\-\.$uscore]+$/)
+	   && ($name !~ /.\*/ || $bind_version >= 9) # "*" can be only the first
 						    # char, for bind 8
-	   && ($_[0] !~ /\*[^\.]/))	# a "." must always follow "*"
-	 : ($_[0] =~ /^[\A-Za-z0-9\-\.$uscore]+$/))
-	&& ($_[0] !~ /\.\./)		# no ".." inside
-	&& ($_[0] !~ /^\../)		# no "." at the beginning
-	&& ($_[0] !~ /^\-/)		# no "-" at the beginning
-	&& ($_[0] !~ /\-$/)		# no "-" at the end
-	&& ($_[0] !~ /\.\-/)		# no ".-" inside
-	&& ($_[0] !~ /\-\./)		# no "-." inside
-	&& ($_[0] !~ /\.[0-9]+\.$/));	# last label in FQDN may not be
+	   && ($name !~ /\*[^\.]/))	# a "." must always follow "*"
+	 : ($name =~ /^[\A-Za-z0-9\-\.$uscore]+$/))
+	&& ($name !~ /\.\./)		# no ".." inside
+	&& ($name !~ /^\../)		# no "." at the beginning
+	&& ($name !~ /^\-/)		# no "-" at the beginning
+	&& ($name !~ /\-$/)		# no "-" at the end
+	&& ($name !~ /\.\-/)		# no ".-" inside
+	&& ($name !~ /\-\./)		# no "-." inside
+	&& ($name !~ /\.[0-9]+\.$/));	# last label in FQDN may not be
 					# purely numeric
 }
 
@@ -636,7 +760,7 @@ sub valemail
 {
 return $_[0] eq "." ||
        $_[0] =~ /^[A-Za-z0-9\.\-]+$/ ||
-       $_[0] =~ /(.*)\@(.*)/ && 
+       $_[0] =~ /(\S*)\@(\S*)/ && 
        &valdnsname($2, 0, ".") && 
        $1 =~ /[a-z][\w\-\.$uscore]+/i;
 }
@@ -645,18 +769,21 @@ return $_[0] eq "." ||
 # If a path does not start with a /, prepend the base directory
 sub absolute_path
 {
-if ($_[0] =~ /^([a-zA-Z]:)?\//) { return $_[0]; }
-return &base_directory()."/".$_[0];
+my ($path) = @_;
+if ($path =~ /^([a-zA-Z]:)?\//) {
+	return $path;
+	}
+return &base_directory()."/".$path;
 }
 
 # parse_spf(text, ...)
 # If some text looks like an SPF TXT record, return a parsed hash ref
 sub parse_spf
 {
-my $txt = join(" ", @_);
+my $txt = join("", @_);
 if ($txt =~ /^v=spf1/) {
-	local @w = split(/\s+/, $txt);
-	local $spf = { };
+	my @w = split(/\s+/, $txt);
+	my $spf = { };
 	foreach my $w (@w) {
 		$w = lc($w);
 		if ($w eq "a" || $w eq "mx" || $w eq "ptr") {
@@ -699,35 +826,93 @@ return undef;
 # into multiple quoted strings.
 sub join_spf
 {
-local ($spf) = @_;
-local @rv = ( "v=spf1" );
+my ($spf) = @_;
+my @rv = ( "v=spf1" );
 foreach my $s ("a", "mx", "ptr") {
 	push(@rv, $s) if ($spf->{$s});
 	}
 foreach my $s ("a", "mx", "ip4", "ip6", "ptr", "include", "exists") {
-	foreach my $v (@{$spf->{"$s:"}}) {
-		push(@rv, "$s:$v");
+	if ($spf->{"$s:"}) {
+		foreach my $v (@{$spf->{"$s:"}}) {
+			push(@rv, "$s:$v");
+			}
 		}
 	}
-push(@rv, @{$spf->{'other'}});
-if ($spf->{'all'} == 3) { push(@rv, "-all"); }
-elsif ($spf->{'all'} == 2) { push(@rv, "~all"); }
-elsif ($spf->{'all'} == 1) { push(@rv, "?all"); }
-elsif ($spf->{'all'} eq '0') { push(@rv, "all"); }
+if ($spf->{'other'}) {
+	push(@rv, @{$spf->{'other'}});
+	}
 foreach my $m ("redirect", "exp") {
 	if ($spf->{$m}) {
 		push(@rv, $m."=".$spf->{$m});
 		}
 	}
-local @rvwords;
-local $rvword;
+if ($spf->{'all'} == 3) { push(@rv, "-all"); }
+elsif ($spf->{'all'} == 2) { push(@rv, "~all"); }
+elsif ($spf->{'all'} == 1) { push(@rv, "?all"); }
+elsif ($spf->{'all'} eq '0') { push(@rv, "all"); }
+my @rvwords;
+my $rvword = "";
+while(@rv) {
+	my $w = shift(@rv);
+	if (length($rvword)+length($w)+1 >= 255) {
+		$rvword .= " ";
+		push(@rvwords, $rvword);
+		$rvword = "";
+		}
+	$rvword .= " " if ($rvword);
+	$rvword .= $w;
+	}
+push(@rvwords, $rvword);
+return join("\" \"", @rvwords);
+}
+
+# parse_dmarc(text, ...)
+# If some text looks like an DMARC TXT record, return a parsed hash ref
+sub parse_dmarc
+{
+my $txt = join(" ", @_);
+if ($txt =~ /^v=dmarc1/i) {
+	my @w = split(/;\s*/, $txt);
+	my $dmarc = { };
+	foreach my $w (@w) {
+		$w = lc($w);
+		if ($w =~ /^(v|pct|ruf|rua|p|sp|adkim|aspf|fo|rf|ri)=(\S+)$/i) {
+			$dmarc->{$1} = $2;
+			}
+		else {
+			push(@{$dmarc->{'other'}}, $w);
+			}
+		}
+	return $dmarc;
+	}
+return undef;
+}
+
+# join_dmarc(&dmarc)
+# Converts a DMARC record structure to a string, designed to be inserted into
+# quotes in a TXT record. If it is longer than 255 bytes, it will be split
+# into multiple quoted strings.
+sub join_dmarc
+{
+my ($dmarc) = @_;
+my @rv = ( "v=DMARC1" );
+foreach my $s ("p", "pct", "ruf", "rua", "sp", "adkim", "aspf", "fo", "rf", "ri") {
+	if ($dmarc->{$s} && $dmarc->{$s} ne '') {
+		push(@rv, $s."=".$dmarc->{$s});
+		}
+	}
+if ($dmarc->{'other'}) {
+	push(@rv, @{$dmarc->{'other'}});
+	}
+my @rvwords;
+my $rvword = "";
 while(@rv) {
 	my $w = shift(@rv);
 	if (length($rvword)+length($w)+1 >= 255) {
 		push(@rvwords, $rvword);
 		$rvword = "";
 		}
-	$rvword .= " " if ($rvword);
+	$rvword .= "; " if ($rvword);
 	$rvword .= $w;
 	}
 push(@rvwords, $rvword);
@@ -739,18 +924,19 @@ return join("\" \"", @rvwords);
 # with quoting if needed
 sub join_record_values
 {
-local ($r) = @_;
+my ($r) = @_;
 if ($r->{'type'} eq 'SOA') {
 	# Multiliple lines, with brackets
-	local $v = $r->{'values'};
+	my $v = $r->{'values'};
 	return "$v->[0] $v->[1] (\n\t\t\t$v->[2]\n\t\t\t$v->[3]\n".
 	       "\t\t\t$v->[4]\n\t\t\t$v->[5]\n\t\t\t$v->[6] )";
 	}
 else {
 	# All one one line
-	local @rv;
+	my @rv;
 	foreach my $v (@{$r->{'values'}}) {
-		push(@rv, $v =~ /\s/ ? "\"$v\"" : $v);
+		push(@rv, $v =~ /\s|;/ || $r->{'type'} eq 'TXT' ?
+				"\"$v\"" : $v);
 		}
 	return join(" ", @rv);
 	}
@@ -760,8 +946,8 @@ else {
 # Given an old serial number, returns a new one using the configured method
 sub compute_serial
 {
-local ($old) = @_;
-if ($config{'soa_style'} == 1 && $old =~ /^(\d{8})(\d\d)$/) {
+my ($old) = @_;
+if ($config{'soa_style'} == 1 && $old =~ /^(\d{8})(\d+)$/) {
 	if ($1 >= &date_serial()) {
 		if ($2 >= 99) {
 			# Have to roll over to next day
@@ -779,7 +965,7 @@ if ($config{'soa_style'} == 1 && $old =~ /^(\d{8})(\d\d)$/) {
 	}
 elsif ($config{'soa_style'} == 2) {
 	# Unix time
-	local $rv = time();
+	my $rv = time();
 	while($rv <= $old) {
 		$rv = $old + 1;
 		}
@@ -795,14 +981,14 @@ else {
 # Make a short name like foo a fully qualified name like foo.domain.com.
 sub convert_to_absolute
 {
-local ($name, $origin) = @_;
+my ($name, $origin) = @_;
 if ($name eq $origin ||
     $name =~ /\.\Q$origin\E$/) {
 	# Name already ends in domain name - add . automatically, so we don't
 	# re-append the domain name.
 	$name .= ".";
 	}
-local $rv = $name eq "" ? "$origin." :
+my $rv = $name eq "" ? "$origin." :
 	    $name eq "@" ? "$origin." :
 	    $name !~ /\.$/ ? "$name.$origin." : $name;
 $rv =~ s/\.+$/\./;
@@ -814,16 +1000,18 @@ return $rv;
 # If absolute is 1, the path is made absolute. If 2, it is also un-chrooted
 sub get_zone_file
 {
-local ($z, $abs) = @_;
-local $fn;
+my ($z, $abs) = @_;
+$abs ||= 0;
+my $fn;
 if ($z->{'members'}) {
-	local $file = &find("file", $z->{'members'});
+	my $file = &find("file", $z->{'members'});
 	return undef if (!$file);
 	$fn = $file->{'values'}->[0];
 	}
 else {
 	$fn = $z->{'file'};
 	}
+return undef if (!$fn);
 if ($abs) {
 	$fn = &absolute_path($fn);
 	}
@@ -834,24 +1022,27 @@ return $fn;
 }
 
 # get_dnskey_record(&zone|&zonename, [&records])
-# Returns the DNSKEY record for some domain, or undef if none
+# Returns the DNSKEY record(s) for some domain, or undef if none
 sub get_dnskey_record
 {
-local ($z, $recs) = @_;
-if (!$recs) {
-	# Need to get zone file and thus records
-	local $fn = &get_zone_file($z);
-	$recs = [ &read_zone_file($fn, $dom) ];
-	}
-# Find the record
-local $dom = $z->{'members'} ? $z->{'values'}->[0] : $z->{'name'};
-foreach my $r (@$recs) {
-	if ($r->{'type'} eq 'DNSKEY' &&
-	    $r->{'name'} eq $dom.'.') {
-		return $r;
+my ($z, $recs) = @_;
+my $dom = $z->{'members'} ? $z->{'values'}->[0] : $z->{'name'};
+my @rv;
+if ($dom) {
+	if (!$recs) {
+		# Need to get zone file and thus records
+		my $fn = &get_zone_file($z);
+		$recs = [ &read_zone_file($fn, $dom) ];
+		}
+	# Find the record
+	foreach my $r (@$recs) {
+		if ($r->{'type'} eq 'DNSKEY' &&
+		    $r->{'name'} eq $dom.'.') {
+			push(@rv, $r);
+			}
 		}
 	}
-return undef;
+return wantarray ? @rv : $rv[0];
 }
 
 # record_id(&r)
@@ -860,7 +1051,8 @@ sub record_id
 {
 my ($r) = @_;
 return $r->{'name'}."/".$r->{'type'}.
-       (uc($r->{'type'}) eq 'SOA' ? '' : '/'.join('/', @{$r->{'values'}}));
+       (uc($r->{'type'}) eq 'SOA' || !$r->{'values'} ? '' :
+		'/'.join('/', @{$r->{'values'}}));
 }
 
 # find_record_by_id(&recs, id, index)
@@ -886,15 +1078,15 @@ else {
 # Returns the DNSKEY recordset for some domain, or an empty array if none 
 sub get_dnskey_rrset
 {
-	local ($z, $recs) = @_;
-	local @rv = ();
+	my ($z, $recs) = @_;
+	my @rv = ();
+	my $dom = $z->{'members'} ? $z->{'values'}->[0] : $z->{'name'};
 	if (!$recs) {
 		# Need to get zone file and thus records
 		my $fn = &get_zone_file($z);
 		$recs = [ &read_zone_file($fn, $dom) ];
 	}
 	# Find the record
-	local $dom = $z->{'members'} ? $z->{'values'}->[0] : $z->{'name'};
 	foreach my $r (@$recs) {
 		if ($r->{'type'} eq 'DNSKEY' &&
 			$r->{'name'} eq $dom.'.') {
@@ -904,5 +1096,17 @@ sub get_dnskey_rrset
 	return @rv;
 }
 
+# is_raw_format_records(file)
+# Checks if a zone file is in BIND's new raw or text format
+sub is_raw_format_records
+{
+my ($file) = @_;
+open(my $RAW, "<", $file) || return 0;
+my $buf;
+read($RAW, $buf, 3);
+close($RAW);
+return $buf eq "\0\0\0";
+}
+
 1;
 
